"""Coil_module."""
# Nov 19, 2022, Ingo Rehberg
import timeit
import numpy as np
from scipy.special import ellipe, ellipk  # ignore err of Spyder code analysis


class CurrentLoop():
    """A circular current loop and its field."""

    def __init__(self, radius=10, phi=0):
        self.__radius = radius
        self.radius_sq = self.__radius*self.__radius
        self.__phi = phi
        self.cos_pr = np.cos(np.pi/180.*self.__phi)
        self.sin_pr = np.sin(np.pi/180.*self.__phi)
        self.x_d = None  # for internal use during rotation
        self.z_d = None  # for internal use during rotation

    @property
    def radius(self):
        """Return the radius of the circular current loop."""
        return self.__radius

    @radius.setter
    def radius(self, rad):
        if rad > 0:
            self.__radius = rad
        else:
            print('radius must be > 0, taking absolute value')
            self.__radius = -rad
        self.radius_sq = self.__radius*self.__radius

    @property
    def phi(self):
        """Return the angle of the circular current loop."""
        return self.__phi

    @phi.setter
    def phi(self, angle):
        self.__phi = angle
        self.cos_pr = np.cos(np.pi/180.*self.__phi)
        self.sin_pr = np.sin(np.pi/180.*self.__phi)

    def turn(self, x_i, z_i):
        """Turn the coordinate system."""
        self.x_d = x_i*self.cos_pr + z_i*self.sin_pr
        self.z_d = -x_i*self.sin_pr + z_i*self.cos_pr

    def turn_back(self, x_i, z_i):
        """Turn the field."""
        return x_i*self.cos_pr - z_i*self.sin_pr, \
            x_i*self.sin_pr + z_i*self.cos_pr

    def h_field(self, x_i, y_d, z_i):
        """Return the field at position x,y,z."""
        self.turn(x_i, z_i)
        rho2 = np.clip(y_d*y_d + self.x_d*self.x_d, 1.e-15, None)  # avoid 0
        rsqu = rho2 + self.z_d*self.z_d
        alfa2 = self.radius**2 + rsqu - 2*self.radius*np.sqrt(rho2)
        beta2 = alfa2 + 4*self.radius*np.sqrt(rho2)
        k = 1 - alfa2/beta2  # k = np.sqrt(k), different from maple!
        he_e = ellipe(k)
        he_k = ellipk(k)
        term_0 = 1/2*self.z_d*(
            (self.radius_sq + rsqu)*he_e
            - alfa2*he_k)/(alfa2*np.pi*np.sqrt(beta2)*rho2)
        h_x = term_0 * self.x_d
        h_y = term_0 * y_d
        h_z = 1/2 * ((self.radius_sq - rsqu)*he_e +
                     alfa2*he_k)/(alfa2*np.pi*np.sqrt(beta2))
        hx_d, hz_d = self.turn_back(h_x, h_z)
        return np.sqrt(hx_d*hx_d + h_y*h_y + hz_d*hz_d), hx_d, h_y, hz_d


    def h_vec(self, x_i, y_d, z_i):
        """Return the field at position x,y,z."""
        self.turn(x_i, z_i)
        rho2 = np.clip(y_d*y_d + self.x_d*self.x_d, 1.e-15, None)  # avoid 0
        rsqu = rho2 + self.z_d*self.z_d
        alfa2 = self.radius**2 + rsqu - 2*self.radius*np.sqrt(rho2)
        beta2 = alfa2 + 4*self.radius*np.sqrt(rho2)
        k = 1 - alfa2/beta2  # k = np.sqrt(k), different from maple!
        he_e = ellipe(k)
        he_k = ellipk(k)
        term_0 = 1/2*self.z_d*(
            (self.radius_sq + rsqu)*he_e
            - alfa2*he_k)/(alfa2*np.pi*np.sqrt(beta2)*rho2)
        h_x = term_0 * self.x_d
        h_y = term_0 * y_d
        h_z = 1/2 * ((self.radius_sq - rsqu)*he_e +
                     alfa2*he_k)/(alfa2*np.pi*np.sqrt(beta2))
        hx_d, hz_d = self.turn_back(h_x, h_z)
        return np.array([hx_d, h_y, hz_d])

class HelmholtzPair():
    """A circular current loop and its field."""

    def __init__(self, radius=1, phi=0, distance=1):
        self.__distance = distance
        self.cur_l = CurrentLoop(radius=radius, phi=phi)

    @property
    def distance(self):
        """Return the distance between the pairs."""
        return self.__distance

    @distance.setter
    def distance(self, dist):
        if dist >= 0:
            self.__distance = dist
        else:
            print('distance should be > 0, taking absolute value')
            self.__distance = -dist

    def h_field(self, x_p, y_p, z_p):
        """Calculate the field from 2 coils seperated by __distance."""
        _, x_1, y_1, z_1 = self.cur_l.h_field(x_p, y_p, z_p+self.__distance/2)
        _, x_2, y_2, z_2 = self.cur_l.h_field(x_p, y_p, z_p-self.__distance/2)
        x_c = (x_1+x_2)/2
        y_c = (y_1+y_2)/2
        z_c = (z_1+z_2)/2
        return np.sqrt(x_c*x_c + y_c*y_c + z_c*z_c), x_c, y_c, z_c


def test_h_field(repetitions):
    """Check the speed of the routines."""
    time = timeit.timeit(
        'helmh.h_field(X_1, 1., Z_1)',
        'from __main__ import X_1, Z_1, helmh', number=repetitions)
    print(f'cl.h_field: {time:.3f}')


if __name__ == '__main__':
    # Demontrate use of the functions and time them.
    helmh = HelmholtzPair(radius=1, distance=0, phi=45.)
    # print(helmh.cur_l.phi, helmh.cur_l.sin_pr, helmh.cur_l.cos_pr)
    RPT = 10**3
    W_P = 1.5
    X_1, Z_1 = np.meshgrid(np.linspace(-W_P, W_P, 100),
                           np.linspace(-W_P, W_P, 100))
    test_h_field(RPT)
