"""Uses numba, without using scipy."""
# Ingo Rehberg Nov 19, 2022
import timeit
import numpy as np
from numba import vectorize, njit, float64


@vectorize([float64(float64)], fastmath=True)
def rd(y_in):
    """Compute incomplete elliptic integral of the second kind, RD(X,Y,Z)."""
    x_n = 0.0
    y_n = y_in
    z_n = 1.0
    sigma = 0.0
    power4 = 1.0
    while True:
        mue = (x_n + y_n + 3.0 * z_n) * 0.2
        xndev = (mue - x_n) / mue
        yndev = (mue - y_n) / mue
        zndev = (mue - z_n) / mue
        epsilon = max(abs(xndev), abs(yndev), abs(zndev))

        if epsilon < 1.0E-03:
            c_1 = 3.0 / 14.0
            c_2 = 1.0 / 6.0
            c_3 = 9.0 / 22.0
            c_4 = 3.0 / 26.0
            e_a = xndev * yndev
            e_b = zndev * zndev
            e_c = e_a - e_b
            e_d = e_a - 6.0*e_b
            e_f = e_d + e_c + e_c
            s_1 = e_d * (- c_1 + 0.25 * c_3 * e_d - 1.5 * c_4 * zndev * e_f)
            s_2 = zndev*(c_2 * e_f + zndev * (- c_3 * e_c + zndev * c_4 * e_a))
            return 3.0*sigma + power4*(1.0 + s_1 + s_2) / (mue * np.sqrt(mue))

        xnroot = np.sqrt(x_n)
        ynroot = np.sqrt(y_n)
        znroot = np.sqrt(z_n)
        lamda = xnroot * (ynroot + znroot) + ynroot * znroot
        sigma = sigma + power4 / (znroot * (z_n + lamda))
        power4 = power4 * 0.25
        x_n = (x_n + lamda) * 0.25
        y_n = (y_n + lamda) * 0.25
        z_n = (z_n + lamda) * 0.25


@vectorize([float64(float64)], fastmath=True)
def rf(y_in):
    """Compute an incomplete elliptic integral of the first kind, RF(X,Y,Z)."""
    x_n = 0.0
    y_n = y_in
    z_n = 1.0

    while True:
        mue = (x_n + y_n + z_n) / 3.0
        xndev = 2.0 - (mue + x_n) / mue
        yndev = 2.0 - (mue + y_n) / mue
        zndev = 2.0 - (mue + z_n) / mue
        epslon = max(abs(xndev), abs(yndev), abs(zndev))

        if epslon < 1.0E-03:
            c_1 = 1.0 / 24.0
            c_2 = 3.0 / 44.0
            c_3 = 1.0 / 14.0
            e_2 = xndev * yndev - zndev * zndev
            e_3 = xndev * yndev * zndev
            return (1.0+(c_1*e_2 - 0.1 - c_2*e_3)*e_2 + c_3 * e_3)/np.sqrt(mue)

        xnroot = np.sqrt(x_n)
        ynroot = np.sqrt(y_n)
        znroot = np.sqrt(z_n)
        lamda = xnroot * (ynroot + znroot) + ynroot * znroot
        x_n = (x_n + lamda) * 0.25
        y_n = (y_n + lamda) * 0.25
        z_n = (z_n + lamda) * 0.25


@njit(parallel=True, fastmath=True)
def turn(x_i, z_i, p_r):
    """Turn the coordinates by p_r given in radian."""
    x_d = x_i*np.cos(p_r) + z_i*np.sin(p_r)
    z_d = -x_i*np.sin(p_r) + z_i*np.cos(p_r)
    return x_d, z_d


@njit(parallel=True, fastmath=True)
def turn_back(h_x, h_z, p_r):
    """Turn the field by p_r given in radian."""
    hx_d = h_x*np.cos(p_r) - h_z*np.sin(p_r)
    hz_d = h_x*np.sin(p_r) + h_z*np.cos(p_r)
    return hx_d, hz_d


@njit(parallel=True, fastmath=True)
def h_circ_cur(x_d, y_d, z_d, radius):
    """Return the field at position x,y,z in the ring coord-system."""
    rho2 = y_d*y_d + x_d*x_d
    rho2 = rho2+1.e-15
    radius_sq = radius*radius
    rsqu = rho2 + z_d*z_d
    h_1 = radius_sq + rsqu
    h_2 = 2*radius*np.sqrt(rho2)
    alfa2 = h_1-h_2
    beta2 = h_1+h_2
    k = 1. - alfa2/beta2  # k = np.sqrt(k), different from maple!
    h_3 = alfa2*np.pi*np.sqrt(beta2)
    eee = rf(1.0 - k)-k*rd(1.0 - k)/3.0  # my_ellipe(k)
    eek = rf(1.0 - k)*alfa2  # my_ellipk(k) * alfa2
    h_1 = 1/2*z_d*(h_1 * eee - eek)/(h_3*rho2)
    h_x = x_d * h_1
    h_y = y_d * h_1
    h_z = 1/2 * ((radius_sq - rsqu)*eee + eek)/h_3
    return h_x, h_y, h_z


@njit(parallel=True, fastmath=True)
def h_field(x_i, y_d, z_i, radius, p_r):
    """Return the field at position x,y,z."""
    x_d, z_d = turn(x_i, z_i, p_r)
    h_x, h_y, h_z = h_circ_cur(x_d, y_d, z_d, radius)
    hx_d, hz_d = turn_back(h_x, h_z, p_r)
    dum = np.sqrt(hx_d**2 + h_y**2 + hz_d**2)
    return dum, hx_d, h_y, hz_d


def test_h_field(repetitions):
    """Check the speed of the routines."""
    time = timeit.timeit(
        'h_field(X_1, 1., Z_1, 1.5, 1.0)',
        'from __main__ import X_1, Z_1, h_field', number=repetitions)
    print(f'cl.h_field: {time:.3f}')


if __name__ == '__main__':
    # import coil_pack_numba.coil_module_numba as cpn
    X_1, Z_1 = np.meshgrid(np.linspace(-1.5, 1.5, 100),
                           np.linspace(-1.5, 1.5, 100))
    test_h_field(1)  # This run includes the "just in time compilation"
    test_h_field(1000)
