"""Coils_toroidal: linear and toroidal coil with finite size effects."""
# coils_toroidal.py, V. 2.0, Nov 12, 2023, Ingo Rehberg
import timeit
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.widgets import Slider, Button, RadioButtons, CheckButtons
from scipy.integrate import ode
USE_NUMBA = False  # set to "False" if you cannot use numba
TEST_SPEED = False  # is useful during program evaluation
# using Numba makes the code approximately 3 times faster with my hardware
if USE_NUMBA:
    import coil_pack.coil_module_numba as cpn
else:
    from coil_pack.coil_module import CurrentLoop
# from matplotlib import use  # use for creating exe, set comment for spyder
# use("TkAgg")                # use for creating exe, set comment for spyder


class Cur:
    """Geometry of the electrical current."""

    n_l = 1        # number of circular current loops
    r_draht = 0.1  # the radius of the wire, no field calculation done inside
    r_tor = 1.6    # radius of the toroid
    r_ring = 1.    # radius of the current ring
    toroid = False
    if not USE_NUMBA:
        loop = CurrentLoop()  # instance for a single current loop
    phi = np.zeros(n_l)       # the angles for the current loops in degree
    phi_rad = np.zeros(n_l)   # the angles for the current loops in rad
    x_pos = np.zeros(n_l)     # x-position of the center of the current loop
    z_pos = np.zeros(n_l)     # z-position of the center of the current loop

    @staticmethod
    def set_ring_geo():
        """toroid."""
        Cur.phi = np.zeros(Cur.n_l)
        for i in range(Cur.n_l):
            Cur.phi[i] = i*360./Cur.n_l  # in degree
        Cur.phi_rad = Cur.phi * np.pi/180.
        Cur.x_pos = Cur.r_tor*np.cos(Cur.phi_rad)
        Cur.z_pos = Cur.r_tor*np.sin(Cur.phi_rad)

    @staticmethod
    def set_cyl_geo():
        """Cylindrical coil."""
        Cur.phi = np.zeros(Cur.n_l)
        Cur.phi[0] = Pl.phi_neu.val if Cur.n_l == 1 else 0

        Cur.phi_rad = Cur.phi * np.pi/180.
        Cur.x_pos = np.zeros(Cur.n_l)
        Cur.z_pos = np.zeros(Cur.n_l)
        for i in range(Cur.n_l):
            Cur.z_pos[i] = Cur.z_start()+i*Cur.distance()

    @staticmethod
    def set_toroid():
        """Set paramters for displaying a toroidal coil."""
        Cur.toroid = True
        Cur.set_ring_geo()
        Cur.r_ring = 0.4
        if not USE_NUMBA:
            Cur.loop.radius = Cur.r_ring
            Cur.loop.phi = Cur.phi[0]
        Pl.cl = 4*h_spule(Cur.r_tor, 0.1, 0.)[0]  # clip level

    @staticmethod
    def distance():
        """Distance between the coil rings."""
        if Cur.n_l == 2:
            return 2*np.abs(Pl.d_neu.val)
        return 0 if Cur.n_l == 1 else 2*1.5/(Cur.n_l-1)

    @staticmethod
    def z_start():
        """Return position lowest ring."""
        if Cur.n_l == 2:
            return -np.abs(Pl.d_neu.val)
        return 0 if Cur.n_l == 1 else -1.5

    @staticmethod
    def set_solenoid():
        """Set paramters for displaying a linear coil."""
        Cur.toroid = False
        Cur.r_ring = 1.
        Cur.set_cyl_geo()
        if USE_NUMBA:
            Pl.cl = cpn.h_field(0.99*Cur.r_ring, 0., 0.,
                                Cur.r_ring, Cur.phi[0])[0]  # clip level
        else:
            Cur.loop.radius = Cur.r_ring
            Cur.loop.phi = Cur.phi[0]
            Pl.cl = Cur.loop.h_field(0.99*Cur.r_ring, 0., 0.)[0]

    @staticmethod
    def get_coil():
        """Get points where the wires cross the plane."""
        pcx = np.zeros(Cur.n_l)
        pcy = np.zeros(Cur.n_l)
        ncx = np.zeros(Cur.n_l)
        ncy = np.zeros(Cur.n_l)
        if Cur.toroid:
            r_i = Cur.r_tor - Cur.r_ring
            r_a = Cur.r_tor + Cur.r_ring
            pcx = r_i*np.cos(Cur.phi_rad)
            pcy = r_i*np.sin(Cur.phi_rad)
            ncx = r_a*np.cos(Cur.phi_rad)
            ncy = r_a*np.sin(Cur.phi_rad)
        else:
            for i in range(Cur.n_l):
                # y_p = Cur.z_start()+i*Cur.distance()
                pcx[i] = -Cur.r_ring*np.cos(Cur.phi_rad[i])
                pcy[i] = Cur.z_pos[i]-Cur.r_ring*np.sin(Cur.phi_rad[i])
                ncx[i] = Cur.r_ring*np.cos(Cur.phi_rad[i])
                ncy[i] = Cur.z_pos[i]+Cur.r_ring*np.sin(Cur.phi_rad[i])
        return pcx, pcy, ncx, ncy

    @staticmethod
    def mindist(x_p, z_p):
        """Get the minimal distance from a point to a wire crossing point."""
        pcx, pcz, ncx, ncz = Cur.get_coil()  # position of +,- points
        rmin = pow(x_p-pcx[0], 2) + pow(z_p-pcz[0], 2)
        for i, pcxi in enumerate(pcx):
            rmin = np.minimum(rmin, (pow(x_p-pcxi, 2) + pow(z_p-pcz[i], 2)))
            rmin = np.minimum(rmin, (pow(x_p-ncx[i], 2) + pow(z_p-ncz[i], 2)))
        return np.sqrt(rmin)


class Pl:
    """Plot layout and functions."""

    plotarrow, plotline = False, False

    z_0 = 0.         # value set by slider
    x_0 = 0.0        # value set by slider
    r_0 = Cur.r_tor  # value set by slider
    cl = []          # clip level
    wp = 2.1         # width of the plot surface
    X1, Z1 = np.meshgrid(np.linspace(-wp, wp, 100), np.linspace(-wp, wp, 100))
    x = np.linspace(-wp, wp, 100)
    # y = x*1.
    z = x*1.
    axi = [None]*5
    pos = [0.05, 0.44, 0.5/1.6, 0.5]
    fig1 = plt.figure('solenoidal and toroidal coils, Helmholtz pair',
                      figsize=(16*0.9, 10*0.9))      # ratio 16/10
    axi[0] = plt.axes(pos)  # upper left, 2d
    axi[1] = plt.axes([pos[0]+pos[2], pos[1], 0.1, pos[3]])  # upper right, H_z
    axi[2] = plt.axes([pos[0],        pos[1]-0.2, pos[2], 0.2])  # lower left
    axi[3] = plt.axes([0.58, 0.39, 0.6/1.6, 0.6], projection='3d')  # right
    axi[4] = plt.axes([(pos[0]+pos[2]+0.1), 0.07, 0.5, 0.3])
    x0_neu = Slider(plt.axes([pos[0], pos[1]+pos[3]+0.005, pos[2], 0.02],
                             facecolor='lightgoldenrodyellow'),
                    '$x_0$', -wp, wp, valinit=x_0)

    z0_neu = Slider(plt.axes([pos[0]+pos[2]+0.10, pos[1],  0.02/1.6, pos[3]],
                             facecolor='lightgoldenrodyellow'),
                    '$z_0$',  -wp, wp, orientation="vertical", valinit=z_0)
    r0_neu = Slider(plt.axes([pos[0]+pos[2]+0.115, pos[1],  0.02/1.6, pos[3]],
                             facecolor='lightgoldenrodyellow'),
                    '$r_0$', -wp, wp, orientation="vertical", valinit=r_0)
    d_neu = Slider(plt.axes([pos[0]+pos[2]+0.13, pos[1], 0.02/1.6, pos[3]]),
                   '$d/2$', -wp, wp, orientation="vertical", valinit=wp)
    phi_neu = Slider(plt.axes([pos[0]+pos[2]+0.145, pos[1], 0.02/1.6, pos[3]]),
                     r'$\phi$', -180, 180, orientation="vertical",
                     valinit=Cur.phi[0])

    dec_n = Button(plt.axes([pos[0]+0.08, pos[1]+pos[3]+0.03, 0.02, 0.02]),
                   '<', color='lightgoldenrodyellow', hovercolor='0.975')
    inc_n = Button(plt.axes([pos[0]+pos[2]-0.02-0.08, pos[1]+pos[3]+0.03,
                             0.02, 0.02]),
                   '>', color='lightgoldenrodyellow', hovercolor='0.975')
    geom = RadioButtons(plt.axes([0.1, 0.1, 0.06, 0.08],
                                 facecolor='lightgoldenrodyellow'),
                        ['solenoid', 'toroid'],
                        active=Cur.toroid)

    field_display = CheckButtons(plt.axes([0.25, 0.1, 0.05, 0.08],
                                 facecolor='lightgoldenrodyellow'),
                                 labels=['Lines', 'Arrow'],
                                 actives=[plotline, plotarrow])

    @staticmethod
    def refresh():
        """Redo all plots."""
        for axis in Pl.axi:
            axis.cla()
        titl = ' current loop' if Cur.n_l == 1 else ' current loops'
        Pl.axi[0].set_title(str(Cur.n_l)+titl, pad=22)
        _ = Cur.set_toroid() if Cur.toroid else Cur.set_solenoid()
        Pl.plot_h_contour(Pl.axi[0])
        if Pl.plotline:
            Pl.plot_field_lines(Pl.axi[0])
        if Pl.plotarrow:
            Pl.plot_arrows(Pl.axi[0])
        Pl.plot_coil(Pl.axi[0])
        Pl.axi[0].plot([Pl.x_0, Pl.x_0], [Pl.z[0], Pl.z[-1]], 'b--')
        Pl.axi[0].set_xlabel(r"x")
        Pl.axi[0].set_ylabel(r"z")
        Pl.axi[0].set_xlim(-Pl.wp, Pl.wp)
        Pl.axi[0].set_ylim(-Pl.wp, Pl.wp)
        if Cur.toroid:
            Pl.plot_h_phi(Pl.axi[0], Pl.axi[2])
            Pl.plot_h(Pl.axi[1])
        else:
            Pl.plot_h_x(Pl.axi[0], Pl.axi[2])
            Pl.plot_h_z(Pl.axi[1])
        Pl.plot_h_3d(Pl.axi[3])
        Pl.plot_h_r(Pl.axi[4])

        Pl.fig1.canvas.draw()

    @staticmethod
    def plot_field_lines(axi):
        """Plot the lines generated by field lines."""
        x_s, y_s = field_lines()
        for x_act, y_act in zip(x_s, y_s):  # plot all field lines
            axi.plot(x_act, y_act, 'b-', lw=0.5)

    @staticmethod
    def plot_arrows(axi):
        """Indicate the direction of the field with arrows."""
        x_1, z_1 = np.meshgrid(np.linspace(-Pl.wp, Pl.wp, 20),
                               np.linspace(-Pl.wp, Pl.wp, 20))
        r_1 = Cur.mindist(x_1, z_1)
        _, h_x, _, h_z = h_spule(x_1, 0., z_1)
        h_x[r_1 < Cur.r_draht] = 0.
        h_z[r_1 < Cur.r_draht] = 0.
        axi.quiver(x_1, z_1, h_x, h_z, color='cyan', pivot='middle')

    @staticmethod
    def plot_coil(axi):
        """Show direction of the electric current by crosses or dots."""
        pcx, pcy, ncx, ncy = Cur.get_coil()
        axi.plot(pcx, pcy, '.g')
        axi.plot(pcx, pcy, 'og', fillstyle='none', markersize=14)
        axi.plot(ncx, ncy, 'xg')
        axi.plot(ncx, ncy, 'og', fillstyle='none', markersize=14)

    @staticmethod
    def plot_h_x(ax1, ax3):
        """Plot the field H in horizontal direction."""
        h_abs = h_spule(Pl.x, 0., Pl.z_0)[0]
        h_abs = np.clip(h_abs, 0., Pl.cl)
        ax3.plot(Pl.x, h_abs, '-r')
        ax3.plot([Pl.x_0, Pl.x_0], [min(h_abs),
                 max(h_abs)], '--', color='blue')
        ax1.plot([Pl.x[0], Pl.x[-1]], [Pl.z_0, Pl.z_0], '--', color='red')
        ax3.set_xlabel(r"$x$")
        ax3.set_ylabel(r"$H$")
        ax3.set_xlim(-Pl.wp, Pl.wp)

    @staticmethod
    def slope_str(xpl, ypl):
        """Plot the field along the radial direction."""
        i = -50
        slope = np.log(ypl[-1] / ypl[i])
        slope /= np.log(xpl[-1]/xpl[i])
        hstr = f'{slope:2.0f}'
        return r'$H\propto x^{' + hstr + '}$'

    @staticmethod
    def plot_h_r(axi):
        """Plot the field along the radial direction."""
        pnum = 101
        xpl = np.geomspace(Pl.wp, 10*Pl.wp, pnum)
        h_abs, h_x, h_y, h_z = h_spule(xpl, 0.0*xpl, 0.0*xpl)
        sto = pnum - 1
        for i in range(pnum):
            if h_abs[i] < 1.e-11:
                sto = i
                break
        if sto < pnum - 1:
            xpl = np.geomspace(Pl.wp, xpl[sto], pnum)
            h_abs, h_x, h_y, h_z = h_spule(xpl, 0.0*xpl, 0.0*xpl)
        axi.loglog(xpl, h_abs, '--g', label='$H(x)$')
        axi.loglog(xpl[-50:-1], h_abs[-50:-1], '-r',
                   label=Pl.slope_str(xpl, h_abs))
        if np.abs(h_x).max() > 1.e-10:
            axi.loglog(xpl, np.abs(h_x), ':b', label='$H_x(x)$')
        if np.abs(h_z).max() > 1.e-10:
            axi.loglog(xpl, np.abs(h_z), '-.m', label='$H_z(x)$')

        axi.set_xlabel(r"distance from the center along $x$")
        axi.set_ylabel(r"magnetic field $H$")
        axi.legend()

    @staticmethod
    def plot_h_phi(ax1, ax3):
        """Plot the field along the azimuthal direction."""
        phi = np.linspace(0, 2*np.pi, 361*2)
        h_abs = h_spule(Pl.r_0*np.cos(phi), 0.0, Pl.r_0*np.sin(phi))[0]
        h_abs = np.clip(h_abs, 0., Pl.cl)
        ax3.plot(phi*180./np.pi, h_abs, '-g')
        ax1.plot(Pl.r_0*np.cos(phi), Pl.r_0*np.sin(phi), 'g--')
        ax3.set_xlabel(r'azimuthal angle $\phi$')
        ax3.set_ylabel(r"$H$")
        ax3.set_xlim(0, 360)

    @staticmethod
    def plot_h_z(ax2):
        """Plot the field along the vertical direction."""
        h_z = h_spule(Pl.z*0+Pl.x_0, 0.001, Pl.z)[3]
        h_z = np.clip(h_z, -1.*Pl.cl, Pl.cl)
        ax2.plot(h_z, Pl.z, '-b')
        ax2.plot([min(h_z), max(h_z)], [Pl.z_0, Pl.z_0], '--', color='red')
        ax2.set_xlabel(r"$H_\mathrm{z}$")
        ax2.set_ylim(-Pl.wp, Pl.wp)

    @staticmethod
    def plot_h(ax2):
        """Plot abs(H-field) along the vertical direction."""
        h_abs = h_spule(Pl.z*0.0+Pl.x_0, 0.001, Pl.z)[0]
        h_abs = np.clip(h_abs, -1.*Pl.cl, Pl.cl)
        ax2.plot(h_abs, Pl.z, '-b')
        # ax2.plot([min(h), max(h)], [z_0, z_0], '--', color='red')
        ax2.set_xlabel(r"$H$")
        ax2.set_ylim(-Pl.wp, Pl.wp)

    @staticmethod
    def plot_h_contour(ax1):
        """Plot abs(H-field) in the x,z-plane."""
        h_1 = h_spule(Pl.X1, 0., Pl.Z1)[0]
        # H1 = np.clip(H1, 0, cl)
        r_1 = Cur.mindist(Pl.X1, Pl.Z1)
        h_1[r_1 < Cur.r_draht] = 0.
        ax1.contourf(Pl.X1, Pl.Z1, h_1, cmap='hot_r')

    @staticmethod
    def plot_h_3d(ax4):
        """Plot abs(H-field) above the x,z-plane."""
        h_abs = h_spule(Pl.X1, 0.0, Pl.Z1)[0]
        # H = np.clip(H, 0, cl)
        r_1 = Cur.mindist(Pl.X1, Pl.Z1)
        h_abs[r_1 < Cur.r_draht] = 0.
        h_abs[r_1 < Cur.r_draht] = h_abs.max()
        ax4.plot_surface(Pl.X1, Pl.Z1, h_abs, cmap='hot_r')
        ax4.set_xlabel(r"$x$")
        ax4.set_ylabel(r"$z$")
        ax4.set_zlabel(r"$H$")


def h_spule(x_p, y_p, z_p):
    """Field of a circular coil, attempts to reduce numerical problems."""
    sav = np.array([[np.zeros_like(x_p)]*3]*Cur.n_l)
    for i in range(0, Cur.n_l):
        if USE_NUMBA:
            _, sav[i, 0], sav[i, 1], sav[i, 2] =\
                cpn.h_field(x_p-Cur.x_pos[i], y_p, z_p-Cur.z_pos[i],
                            Cur.r_ring, Cur.phi_rad[i])

        else:
            Cur.loop.phi = Cur.phi[i]
            sav[i] = Cur.loop.h_vec(x_p-Cur.x_pos[i], y_p, z_p-Cur.z_pos[i])
    acc = np.sum(sav, axis=0)
    return np.sqrt(pow(acc[0], 2) + pow(acc[1], 2) + pow(acc[2], 2)
                   ), acc[0], acc[1], acc[2]


def h_dir(_, pos_xz):
    """Give direction of the field."""
    _, h_x, _, h_z = h_spule(pos_xz[0], 0., pos_xz[1])
    norm = np.sqrt(h_x*h_x+h_z*h_z)
    return [h_x/norm, h_z/norm]


def field_lines():
    """Create field lines for later plots."""
    x_s, y_s = [], []
    if Cur.toroid:
        line_s = Cur.r_tor - Cur.r_ring + 0.1
        line_e = Cur.r_tor + Cur.r_ring - 0.1
        line_n = 8
    else:
        line_s = -0.899*Cur.r_ring
        line_e = 0.899*Cur.r_ring
        line_n = 8
    step = 0.1  # step width of the integration
    prec2 = pow(step*2, 2)  # precision for end detection
    for alpha in np.linspace(line_s, line_e, line_n):
        path = ode(h_dir)
        path.set_integrator('vode')
        # r.set_f_params(Cur.n_l)
        trace_x = [alpha]
        trace_y = [0]
        path.set_initial_value([trace_x[0], trace_y[0]], 0)
        while path.successful() and path.t < 1000*Pl.wp:
            path.integrate(path.t+step)
            trace_x.append(path.y[0])
            trace_y.append(path.y[1])
            dist = pow(trace_x[-1]-trace_x[0], 2) \
                + pow(trace_y[-1]-trace_y[0], 2)
            if dist < prec2 and path.t > 1.:
                trace_x.append(trace_x[0])  # closes the field line
                trace_y.append(trace_y[0])  # closes the field line
                break
        x_s.append(trace_x)
        y_s.append(trace_y)
    return x_s, y_s


class MW:
    """My widgets."""

    @staticmethod
    def new_field_display(_):
        """Set up to 2 display options."""
        Pl.plotline, Pl.plotarrow = Pl.field_display.get_status()
        Pl.refresh()

    @staticmethod
    def new_geom(label):
        """Switch between 2 different geometry options."""
        Cur.toroid = label == 'toroid'
        MW.set_geom_par()

    @staticmethod
    def connect_widgets():
        """Connect actions with widgets."""
        Pl.dec_n.on_clicked(MW.decrease)
        Pl.inc_n.on_clicked(MW.increase)
        Pl.x0_neu.on_changed(MW.update_x0)
        Pl.z0_neu.on_changed(MW.update_z0)
        Pl.r0_neu.on_changed(MW.update_r0)
        Pl.d_neu.on_changed(MW.update_d)
        Pl.phi_neu.on_changed(MW.update_phi)
        Pl.geom.on_clicked(MW.new_geom)
        Pl.field_display.on_clicked(MW.new_field_display)

    @staticmethod
    def set_geom_par():
        """Set paramters for the geometry."""
        _ = Cur.set_toroid() if Cur.toroid else Cur.set_solenoid()
        Pl.r0_neu.ax.set_visible(Cur.toroid)
        Pl.z0_neu.ax.set_visible(not Cur.toroid)
        Pl.d_neu.ax.set_visible(Cur.n_l == 2 and not Cur.toroid)
        Pl.phi_neu.ax.set_visible(Cur.n_l == 1 and not Cur.toroid)
        Pl.refresh()

    @staticmethod
    def decrease(_):
        """Decrease the number of circular current loops."""
        Cur.n_l = max(Cur.n_l-1, 1)
        MW.set_geom_par()

    @staticmethod
    def increase(_):
        """Increase the number of circular current loops."""
        Cur.n_l = min(Cur.n_l+1, 40)
        MW.set_geom_par()

    @staticmethod
    def update_z0(_):
        """Set new value for the height of the horizontal cut."""
        Pl.z_0 = Pl.z0_neu.val
        Pl.refresh()

    @staticmethod
    def update_x0(_):
        """Set new value for the position of the vertical cut."""
        Pl.x_0 = Pl.x0_neu.val
        Pl.refresh()

    @staticmethod
    def update_r0(_):
        """Set new radius for the azimuthal plot."""
        Pl.r_0 = max(abs(Pl.r0_neu.val), 0.05)
        Pl.refresh()

    @staticmethod
    def update_d(_):
        """Set the distance d according to slider."""
        # PO.helmh.distance = PO.d_neu.val
        Pl.refresh()

    @staticmethod
    def update_phi(_):
        """Set the orientation of the coil according to slider."""
        # Cur.loop.phi = Pl.phi_neu.val
        Cur.phi[0] = Pl.phi_neu.val
        Pl.refresh()


class Sp:
    """Speed test."""

    RPT = 10**1  # number of repetions for comparing the excution speed
    X_1, Z_1 = np.meshgrid(np.linspace(-1954, 1954, 1000),
                           np.linspace(-1954, 1954, 1000))

    @staticmethod
    def test_h_field(repetitions):
        """Check the speed of the routines."""
        d_t = timeit.timeit(
            'cpn.h_field(Sp.X_1, 0., Sp.Z_1, 1., 1.)',
            'from __main__ import Sp, cpn', number=repetitions)
        print(f'cl.h_field: {d_t:.3f}')

    @staticmethod
    def test_h_field_class(repetitions):
        """Check the speed of the routines from the CurrentLoop class."""
        d_t = timeit.timeit(
            'Cur.loop.h_field(Sp.X_1, 0., Sp.Z_1)',
            'from __main__ import Cur, Sp, Pl', number=repetitions)
        print(f'CurrentLoop.h_field: {d_t:.3f}')

    @staticmethod
    def speedtest():
        """Can be used to check the performance using numba."""
        if USE_NUMBA:
            Sp.test_h_field(1)  # this first call includes the compilation time
            Sp.test_h_field(Sp.RPT)
        else:
            Sp.test_h_field_class(Sp.RPT)


if TEST_SPEED:
    Sp.speedtest()

MW.connect_widgets()
MW.set_geom_par()
Pl.refresh()  # plot or replot all plots
plt.show()
