"""Fit two H_z measurements in r and z direction to Helmholtz-theory."""
# combo_fit_helmholtz.py", Ingo Rehberg Jul. 12, 2023
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
from coil_pack.coil_module import HelmholtzPair

z1 = np.array([0, 1, 2, 3, 4, 5, 6, 8.3, 9.3, 10.3, 11.3, 12.3, 13.3, 14.3,
              15.3, 16.3, 17.3, 18.3, 19.3, 20.3])

bx1 = np.array([1, 1, 1, 1, 1, 1, 1, 0.88, 0.83, 0.8, 0.75, 0.7, 0.64, 0.58,
                0.54, 0.5, 0.45, 0.41, 0.37, 0.33])

x2 = np.array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9,
               10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
               20, 21, 22, 23, 24, 25])
bx2 = np.array([1, 1, 1, 1, 1, 1, 1, 0.95, 0.93, 0.89, 0.82, 0.74, 0.62, 0.48,
                0.33, 0.22, 0.1, 0.01, -0.04, -0.08, -0.11, -0.12, -0.12,
                -0.12, -0.12, -0.12])

combo_x = np.append(z1, x2)
combo_y = np.append(bx1, bx2)
helmh = HelmholtzPair(radius=1, distance=1)  # is used by 3 fitting functions


def helm_bz_x(x_abscissa, f0_a, f1_r, f2_d):
    """B_z(x) for the Helmholtz pair, Fit to this function."""
    helmh.distance = f2_d
    helmh.cur_l.radius = f1_r
    b_z = helmh.h_field(x_abscissa, 0., 0.)[3]
    return np.array(b_z)*f0_a


def helm_bz_z(z_abscissa, f0_a, f1_r, f2_d):
    """B_z(z) for the Helmholtz pair, Fit to this function."""
    helmh.distance = f2_d
    helmh.cur_l.radius = f1_r
    b_z = helmh.h_field(0., 0., z_abscissa)[3]
    return np.array(b_z)*f0_a


def combined_function(combo_data, f0_a, f1_r, f2_d):
    """Single data reference passed in."""
    extract1 = combo_data[:len(z1)]  # extract first data
    extract2 = combo_data[len(z1):]  # second data
    result1 = helm_bz_z(extract1, f0_a, f1_r, f2_d)
    result2 = helm_bz_x(extract2, f0_a, f1_r, f2_d)
    return np.append(result1, result2)


def set_my_rc():
    """Adjust to your needs."""
    plt.rc('axes', facecolor='white', edgecolor='black', titlesize=20,
           labelsize=15, grid=False)  # set the format of the plotaxes
    plt.rc('legend', fontsize=12, frameon=True, numpoints=1, labelspacing=0.3,
           columnspacing=0)    # set the format of the legend
    plt.rc('xtick', labelsize=12, top=True, direction='in')
    plt.rc('ytick', labelsize=12, right=True, direction='in')
    plt.rc('font', family="sans-serif", size=12)
    plt.rc('text', usetex=False)


def layout():
    """Layout of the subplots."""
    f_1 = plt.figure(1, figsize=(16*0.7, 10*0.7))  # querformat 16/10
    f_1.canvas.manager.set_window_title('combo_fit_circular loop')
    set_my_rc()

    f_1 = plt.figure(1, figsize=(16*0.7, 10*0.7))  # querformat 16/10
    f_1.canvas.manager.set_window_title(
        'Fit to magnetic field measuremts in a Helmholtz pair of coils')
    set_my_rc()
    a_loc = [None]*3
    a_loc[0] = plt.axes([0.1, 0.59, 0.4, 0.4])  # links,
    a_loc[1] = plt.axes([0.55, 0.59, 0.4, 0.4])  # rechts
    a_loc[2] = plt.axes([0.3, 0.1, 0.4, 0.4], projection='3d')
    return f_1, a_loc


def pl_0(ax1):
    """First plot."""
    ax1.plot(z1, bx1, '.r', label='measurement Dabisch')
    print('single fit 1:',
          curve_fit(helm_bz_z, z1, bx1, p0=(38., 13.72, 13.66),
                    ftol=1.e-15, xtol=1.e-15, gtol=1.e-15)[0])
    z_theory = np.linspace(0., 20., 100)
    label = f'Helmholtz: \n $R$={res0[1]:2.2f} cm \n $d$={res0[2]:2.2f} cm'
    ax1.plot(z_theory, helm_bz_z(z_theory, res0[0], res0[1], res0[2]),
             '-b', label=label)
    ax1.set_xlabel(r'Axial direction $z$ (cm)')
    ax1.set_ylabel(r'Magn. induction $B_\mathrm{z} (mT)$')
    ax1.legend()


def pl_1(ax2):
    """2. plot."""
    ax2.plot(x2, bx2, 'or', mfc='none', label='measurement Dabisch')
    print('single fit 2:', curve_fit(helm_bz_x, x2[1:], bx2[1:],
                                     p0=(38., 13.7, 13.7),
                                     ftol=1.e-15, xtol=1.e-15, gtol=1.e-15)[0])
    x_theory = np.linspace(0., 25., 100)
    label = f'Helmholtz: \n $R$={res0[1]:2.2f} cm \n $d$={res0[2]:2.2f} cm'
    ax2.plot(x_theory, helm_bz_x(x_theory, res0[0], res0[1], res0[2]), '-b',
             label=label)
    ax2.set_xlabel(r'Radial direction $r$ (cm)')
    ax2.set_ylabel(r'')
    ax2.legend()


def pl_2(ax3):
    """3. plot."""
    x_g, z_g = np.meshgrid(np.linspace(-1, 25, 101), np.linspace(-1, 20, 101))
    h_g = res0[0]*helmh.h_field(x_g, 0, z_g)[3]
    h_g = np.clip(h_g, -0.5, 1.5)
    ax3.plot_surface(x_g, z_g, h_g, cmap='seismic', alpha=0.5)
    ax3.scatter(z1*0., z1, bx1, color='g')
    ax3.scatter(x2, x2*0., bx2, color='g')
    ax3.set_xlabel(r"$x$")
    ax3.set_ylabel(r"$z$")
    ax3.set_zlabel(r"$B_\mathrm{z}$")
    fig1.canvas.draw()


res0 = curve_fit(combined_function, combo_x, combo_y, p0=(38., 13.72, 13.66),
                 ftol=1.e-15, xtol=1.e-15, gtol=1.e-15)[0]
print('combo_result:', res0)
fig1, ax_glob = layout()
for ax_num, sub_p in enumerate([pl_0, pl_1, pl_2]):
    sub_p(ax_glob[ax_num])
plt.savefig("combo_helmholtz_fit.png")
plt.show()
